import sys
import os
from fabric.api import *
from fabric.state import output
from fabric.operations import _shell_escape
from fabric.contrib.console import confirm
from fabric.contrib.files import exists

""" Default values """
separator = '-'
confirmation_enabled = False
PROJECT_CONFIGS = 'projects.%s'

class DbAction(object):
    connect = """mysql -u %(db_user)s -p%(db_pass)s -h %(db_host)s"""
    validate_check = ['db_user', 'db_pass', 'db_host', 'db_name'] 
    commands = ''
    
    def __init__(self, env):
        self.env = env
        
    def setup(self):
        pass
    
    def validate(self):
        for var_name in self.validate_check:
            if not self.env.get(var_name, False):
                raise Exception('DbAction doesnt validate. Missing %s'%var_name)
            # TODO: custom exceptions

    def cmd(self):
        return (self.commands % self.env)
    
    def perform(self):
        self.setup()
        self.validate()
        if self.env.get('db_file'):
            self.env.file = self.env.get('db_file')
        return self.cmd()

class DbActionFile(DbAction):
    #validate_check = DbAction.validate_check+['db_file']
    def validate(self):
        super(DbActionFile, self).validate()
        self.env.setdefault('db_file', self.env.get('file'))
        self.env.setdefault('db_file', self.env.get('db_name', 'dump').lower()+".sql")
        
class DbActionClean(DbAction):
    commands = DbAction.connect+""" %(db_name)s """+\
               """ -e "SHOW TABLES;" | grep -v Tables_in | grep -v "+" | gawk '{print "drop table " $1 ";"}' | """+\
               DbAction.connect+""" %(db_name)s """
    
class DbActionExport(DbActionFile):
    commands = """mysqldump --add-drop-table -u %(db_user)s -p%(db_pass)s -h %(db_host)s %(db_name)s > %(db_file)s"""

class DbActionImport(DbActionFile):
    commands = DbAction.connect+""" %(db_name)s < %(db_file)s"""

class DbActionShow(DbAction):
    commands = DbAction.connect+""" -e "SHOW DATABASES;" """

class DbActionCreate(DbAction):
    commands = DbAction.connect+""" -e "CREATE DATABASE IF NOT EXISTS  %(db_name)s;" """

class DbActionGrant(DbAction):    
    commands = DbAction.connect+""" -e "GRANT ALL ON %(db_name_new)s.* TO %(db_user_new)s@%(db_host_new)s IDENTIFIED BY '%(db_pass_new)s';" """,

def c(key):
    """ c = configuration, update environment with the selected config settings """
    if not config.get(key):
        raise Exception("%s config was not found. Did you forget to load it first ?"%key)
    env.update(config.get(key))

def e(**kwargs):
    """ e = environment, update environment with key,value pairs """
    env.update(kwargs)

def q(action_name):
    """
        q = query or database commands
        fab c:localhost c:localdb e:db_name=test q:clean
        fab c:localhost c:localdb e:db_name=test e:db_file=dump.sql q:export
        fab c:localhost c:localdb c:localdbnew e:db_name=test_2 q:create q:grant
        fab e:db_name=test_elgg e:db_file=stuff.sql q:export
        
        fab c:localdb e:db_name=test_elgg q:export
        fab c:localdb e:db_name=test e:db_file=test_elgg.sql q:clean q:import
        
        fab c:localhost c:localdb e:db_name=test_elgg e:db_file=dump.sql q:export l:temp c:dmeh up:e.db_file
    """
    action_class = globals().get( 'DbAction%s'%action_name.capitalize(), False )
    
    if not action_class:
        raise Exception('Unknown database action: %s' % action_name)
    
    action = action_class(env)
    with cd(env.base_dir):
        env.db_result = do(action.perform())

def l(key, conf_path=PROJECT_CONFIGS):
    """ l = load, load configuration """
    current_dir = os.path.dirname(os.path.abspath(__file__))
    if current_dir not in sys.path: 
        sys.path.append( os.path.dirname(os.path.abspath(__file__)) )

    try:
        conf_module = conf_path%key
        config_module = __import__(conf_module, globals(), locals(), ['*'])
        config.update(config_module.config)
        del config_module.config
        for k in dir(config_module):
            globals()[k] = getattr(config_module, k)
    except ImportError:
        raise # what ?

def path_is_absolute(pathname):
    return pathname.startswith('/') or pathname.startswith('~')

def path_is(pathname, parameter='f'):
    """
    check if pathname is either file(f), directory(d) or anything that exists(e)
    """
    cmd = '[ -%s %s ] && echo "yes" || echo "no"'% (parameter, pathname)
    file_info =  do(cmd)
    return True if 'yes' in file_info else False

def path_exists(pathname, local=None):
    return exists(pathname) if (not local and env.get('host_string')) else os.path.exists(pathname)

def path_ask_overwrite(pathname, local=None):
    """
    return True if file doesn't exist or user choose Yes
    return False if user choose No
    """
    if path_exists(pathname, local):
        if not confirm("Found (%s) (%s). Overwrite ?" % (local or env.host, pathname)):
            return False
    return True

def file_setup(file, envname='file', default=None, base_dir=None):
    file = file or env.get(envname) or default # or env.get('last_file')
    if file.startswith('e.') and file[2:] in env:
        file = env.get(file[2:])
        
    if not file:
        raise Exception("No file available.")
    
    if not path_is_absolute(file):
        if not base_dir:
            base_dir = env.base_dir
        file = os.path.join(base_dir, file)
    return file
    
def zp(file=None, dest=None):
    """ 
        zip a source dir or file
        fab zp:temp
        fab e:file=temp,dest=mytemp.zip zp
        
    """
    file = file_setup(file)
    if not path_is(file, 'e'):
        raise Exception("File %s not found."%file)
    dest = file_setup(dest, 'dest', os.path.basename(file)+'.%s.zip' % env.host)

    if path_ask_overwrite(dest):
        with cd(os.path.dirname(file)):
            file = os.path.basename(file)
            do('zip -r %(dest)s %(file)s -x "%(file)s/.git/*"' % locals()) # -P %(zip_pass)s # -x "%(file)s/_private*/*" "%(file)s/sites/*"
    env.file = dest

def zpi(file=None, dest=None):
    """ 
    zip a directory (file = directory)
    chdir into that directory and archive the contents without including the dir itself
    """
    file = file_setup(file)
    if not path_is(file, 'e'):
        raise Exception("File %s not found."%file)
    dest = file_setup(dest, 'dest', os.path.basename(file)+'.%s.zip' % env.host)

    if path_ask_overwrite(dest):
        with cd(file):
            do('zip -r %(dest)s . -x ".git/*"' % locals()) # -P %(zip_pass)s # -x "%(zip_target)s/_private*/*" "%(zip_target)s/sites/*"
    env.file = dest

def uzp(file=None, dest=None):
    """
    fab uzp:filename,dest
    fab e:file=filename uzp:,dest
    """
    file = file_setup(file)
    if not path_is(file, 'e'):
        raise Exception("File %s not found."%file)
    dest = file_setup(dest, 'dest', os.path.dirname(file))
    do("unzip -o %s -d %s"%(file, dest))

def up(source=None, dest=None):
    """
    upload a file to a remote host
    fab l:temp c:dmeh up:fabric.localhost.zip,stuff
    """
    if env.host=='localhost':
        raise Exception("No remote host available.")
    
    source = file_setup( file=source, base_dir=config['localhost'].get('base_dir', os.path.abspath(".")) )
    if not os.path.exists(source):
        raise Exception("File %s not found."%source)
    dest = file_setup( dest, 'dest', env.base_dir)
    
    if path_is(dest, 'd'):
        dest = os.path.join(dest, os.path.basename(source))

    if path_ask_overwrite(dest):
        put(source, dest)
    env.file = dest

def dl(source=None,dest=None):
    """ 
    download a file from remote host
    fab l:temp c:dmeh dl:remotefile.zip
    """
    # TODO: env.last_file should be a stack ?, last_file should depend on env.host
    if env.host=='localhost':
        raise Exception("No remote host available.")
    
    source = file_setup(source)
    if not path_is(source, 'e'):
        raise Exception("File %s not found."%source)
    dest = file_setup( dest, 'dest', config['localhost'].get('base_dir', '.'))

    if os.path.isdir(dest):
        dest = os.path.join(dest, os.path.basename(source))

    if path_ask_overwrite(dest):
        get(source, dest)

def cl(file=None):
    """
    delete a file
    """
    file = file_setup(file)
    if not path_is(file, 'e'):
        raise Exception("File %s not found."%file)
    do("rm -rf %s"%file)

def do(cmd):
    """ run command on remote or localhost """
    print "-------Executing Command-------\n%s\n" % (cmd)
    if confirmation_enabled and not confirm("Are you sure?"):
        abort('Aborted by you.')
    return run(cmd) if env.get('host_string') else local(cmd)

def conf():
    print config
    
""" default configuration """
config = {
  'localhost': {
      'host_string': None,
      'password': None,
      'host': 'localhost',
      'port': 22,
      'base_dir': os.path.abspath("."), # current directory
      #'base_dir': os.path.dirname(os.path.abspath(__file__)),
  }
}

""" the default environment is localhost """
c('localhost')

""" load extra local default settings """
l(PROJECT_CONFIGS%'default', '%s')
#c('localdb')
#c('localdbnew')

